/*
 * Debian menu system -- install-menu
 * install-menus/functions.cc
 *
 * Copyright (C) 1996-2003  Joost Witteveen, 
 * Copyright (C) 2002-2004  Bill Allombert and Morten Brix Pedersen.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License with
 * the Debian GNU/Linux distribution in file /usr/share/common-licenses/GPL;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * Written by Joost Witteveen.
 */
#include <errno.h>
#include "functions.h"
#include "install-menu.h"
#include <stringtoolbox.h>
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <iconv.h>

using std::ostream;
using std::vector;
using std::string;
using std::map;
using std::cerr;
using std::endl;

bool empty_string(const string &s)
{
  if (s.empty() || s == "none")
    return true;
  else
    return false;
}

/**
 * this method translate text
 */
static string 
do_translate( string text, string locale, bool need_recode )
{
	const char *from = text.c_str();
	char *lang = (char *) locale.c_str();
	const char *utf8 = "UTF-8";
	const char *charset;
	char *p = strrchr( lang, '.' );
 
	if ( !p )
		charset = utf8;
	else
	{
		*p = '\0';
		charset = p + 1;
	}
	string tmp_locale = lang;
	tmp_locale += ".";
	tmp_locale += utf8;

	setlocale( LC_ALL, tmp_locale.c_str() );

	const char *to = dgettext( "menu-messages", from );

	if ( need_recode && strcmp( charset, utf8 ) )
	{
		iconv_t id = iconv_open( charset, utf8 );
		if ( (iconv_t) -1 == id )
		{
			cerr << "iconv_open from " << utf8 << " to " << charset << " failed: " << strerror(errno) << endl;
			return "";
		}

		const char *in = to;
		size_t	in_len = strlen(in);
		size_t	out_len = in_len;
		char buf[ out_len + 1 ];
		char *out = buf;

		memset(buf, 0, sizeof buf);
		if ( iconv( id, (char **)(&in), &in_len, &out, &out_len ) == (size_t)-1 )
		{
			iconv_close( id );
			cerr << "iconv from " << utf8 << " to " << charset << " failed for: " << to <<
			    " (" << strerror(errno) << ')' << endl;
			return "";
		}

		iconv_close( id );
		return buf;
	} else
		return to;
}


namespace functions {

ostream &prefix::output(ostream &o, vector<cat_str *> &,
    map<string, string> &)
{
  return o << menumethod->prefix();
}

ostream &shell::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string command = args[0]->soutput(menuentry);
  FILE *status = popen(command.c_str(), "r");

  if (!status)
      throw exceptions::pipeerror_read(command);

  while (!feof(status))
  {
    char tmp[MAX_LINE];
    if (fgets(tmp, MAX_LINE, status) != NULL)
        o << tmp;
  }
  pclose(status);
  return o;
}

ostream &ifroot::output(ostream &o, vector<cat_str *> & args,
    map<string, string> &menuentry)
{
  if(!is_root)
    args[1]->output(o,menuentry);
  else
    args[0]->output(o,menuentry);

  return o;
}

ostream &print::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);

  if (empty_string(s)) {
    cerr << _("Zero-size argument to print function.");
    throw exceptions::informed_fatal();
  }
  return o << s;
}

ostream &ifempty::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);

  if (empty_string(s))
    args[1]->output(o,menuentry);
  return o;
}

ostream &ifnempty::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  if (!empty_string(s))
    args[1]->output(o,menuentry);
  return o;
}

ostream &iffile::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  std::ifstream f(s.c_str());
  if(f)
    args[1]->output(o,menuentry);
  return o;
}

ostream &ifelsefile::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  std::ifstream f(s.c_str());
  if(f)
    args[1]->output(o,menuentry);
  else
    args[2]->output(o,menuentry);
  return o;
}

ostream &catfile::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  std::ifstream f(s.c_str());
  char c;

  while(f && f.get(c))
    o<<c;
  return o;
}
ostream &forall::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  const string &array=args[0]->soutput(menuentry);
  vector<string> vec;
  vector<string>::const_iterator i;
  string s;
  string varname=args[1]->soutput(menuentry);

  break_char(array, vec, ':');

  for(i = vec.begin(); i != vec.end(); ++i){
    menuentry[varname]=(*i);
    s+=args[2]->soutput(menuentry);
  }
  return o<<s;
}

ostream &esc::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o << escape(args[0]->soutput(menuentry),
                     args[1]->soutput(menuentry));
}

ostream &escwith::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o << escapewith(args[0]->soutput(menuentry),
                         args[1]->soutput(menuentry),
                         args[2]->soutput(menuentry));
}

ostream &escfirst::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  string esc=args[1]->soutput(menuentry);
  string t;

  for(string::size_type i=0; i != s.length(); ++i){
    if(!esc.empty() && contains(esc, s[i])) {
      t=s.substr(0,i);
      t+=args[2]->soutput(menuentry);
      t+=s.substr(i);
      break;
    }
    t+=s[i];
  }
  return o<<t;
}
ostream &tolower::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o<<lowercase(args[0]->soutput(menuentry));
}

ostream &toupper::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o << uppercase(args[0]->soutput(menuentry));
}
ostream &replacewith::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o << ::replacewith(args[0]->soutput(menuentry),
                            args[1]->soutput(menuentry),
                            args[2]->soutput(menuentry));
}

ostream &replace::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
    return o << ::replace(args[0]->soutput(menuentry),
                          args[1]->soutput(menuentry),
                          args[2]->soutput(menuentry));
}

ostream &nstring::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int count= stringtoi(args[0]->soutput(menuentry));
  int i;

  for(i=0; i<count; i++)
    o<<args[1]->soutput(menuentry);

  return o;
}
ostream &cppesc::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  return o << ::cppesc(args[0]->soutput(menuentry));
}

ostream &add::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int x=stringtoi(args[0]->soutput(menuentry));
  int y=stringtoi(args[1]->soutput(menuentry));

  return o<<itostring(x+y);
}

ostream &sub::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int x=stringtoi(args[0]->soutput(menuentry));
  int y=stringtoi(args[1]->soutput(menuentry));

  return o<<itostring(x-y);
}

ostream &mult::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int x=stringtoi(args[0]->soutput(menuentry));
  int y=stringtoi(args[1]->soutput(menuentry));

  return o<<itostring(x*y);
}

ostream &div::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int x=stringtoi(args[0]->soutput(menuentry));
  int y=stringtoi(args[1]->soutput(menuentry));
  
  if(y)
    return o<<itostring(x/y);
  else
    return o<<"0";
}

ostream &ifelse::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);

  (!empty_string(s)) ?
    args[1]->output(o,menuentry)
    :
    args[2]->output(o,menuentry);
  return o;
}

ostream &ifeq::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  string t=args[1]->soutput(menuentry);
  if(s==t)
    args[2]->output(o,menuentry);
  
  return o;
}
ostream &ifneq::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  string t=args[1]->soutput(menuentry);
  if(!(s==t))
    args[2]->output(o,menuentry);
  
  return o;
}
ostream &ifeqelse::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  string t=args[1]->soutput(menuentry);
  if(s==t)
    args[2]->output(o,menuentry);
  else
    args[3]->output(o,menuentry);

  return o;
}

ostream &cond_surr::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);

  if(!empty_string(s)){
    args[1]->output(o,menuentry);
    args[0]->output(o,menuentry);
    args[2]->output(o,menuentry);
  }
  return o;
}

ostream &parent::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);

  return o<<string_parent(s);
}

ostream &basename::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  
  return o<<string_basename(s);
}

ostream &stripdir::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string s=args[0]->soutput(menuentry);
  
  return o<<string_stripdir(s);
}

ostream &entrycount::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menuentry[PRIVATE_ENTRYCOUNT_VAR];
}

ostream &entryindex::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menuentry[PRIVATE_ENTRYINDEX_VAR];
}

ostream &firstentry::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int index=stringtoi(menuentry[PRIVATE_ENTRYINDEX_VAR]);

  if(index == 0)
    args[0]->output(o,menuentry);
  return o;
}

ostream &lastentry::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  int index=stringtoi(menuentry[PRIVATE_ENTRYINDEX_VAR]);
  int count=stringtoi(menuentry[PRIVATE_ENTRYCOUNT_VAR]);

  if(index+1 == count)
    args[0]->output(o,menuentry);
  return o;
}

ostream &level::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menuentry[PRIVATE_LEVEL_VAR];
}

ostream &rcfile::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->rcfile();
}
ostream &examplercfile::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->examplercfile();
}
ostream &mainmenutitle::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->mainmenutitle();
}
ostream &rootsection::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->rootsection();
}

ostream &rootprefix::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->rootprefix()->soutput(menuentry);
}

ostream &userprefix::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->userprefix()->soutput(menuentry);
}

ostream &treewalk::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->treewalk();
}
ostream &postoutput::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->postoutput();
}
ostream &preoutput::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  return o<<menumethod->preoutput();
}
ostream &cwd::output(ostream &o, vector<cat_str *> &/*args*/,
    map<string, string> &menuentry)
{
  char buf[300];
  //Bug: getcwd returns NULL if strlen(cwd) > sizeof(buf), so we get
  //     a segfault here.
  return o<<string(getcwd(buf,sizeof(buf)));
}

ostream &translate::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string lang=args[0]->soutput(menuentry);
  string text=args[1]->soutput(menuentry);

  return o << do_translate(text,lang,true);
}

/*
 * this method return translation in UTF-8 (needed for KDE 2.0 and above)
 */
ostream &encode_translate::output(ostream &o, vector<cat_str *> &args,
				 map<string, string> &menuentry){
  string lang=args[0]->soutput(menuentry);
  string text=args[1]->soutput(menuentry);
  return o<<do_translate(text,lang,false);
}

/*
 * this method encode input in locale encoding
 */
ostream &encode::output(ostream &o, vector<cat_str *> &args,
			     map<string, string> &menuentry){
  string text=args[0]->soutput(menuentry);
  string from=args[1]->soutput(menuentry);
  string to=args[2]->soutput(menuentry);

  /* convert to dest charset*/
  iconv_t iconvcontext;
  if ((iconvcontext = iconv_open (to.c_str(), 
				  from.c_str())) == (iconv_t)-1)
  return o<<text;
  
  const char *start = text.c_str();

  size_t srclen = strlen(start);
  char buf[srclen * 4 + 1];
  size_t destlen = sizeof(buf) - 1;
  char *tmp = buf;

  if ( iconv(iconvcontext, (char**)&start, &srclen, &tmp, &destlen) == (size_t)-1 )
    cerr << "iconv from " << from << " to " << to << " failed for: " << text << endl;

  iconv_close(iconvcontext);
  *tmp='\0';
  
  /* free memory */
  return o << buf;
}

/*
 * this method get substring from string
 */
ostream &substr::output(ostream &o, vector<cat_str *> &args,
			   map<string, string> &menuentry){
  int ps, sz;
  string s=args[0]->soutput(menuentry);
  string sps=args[1]->soutput(menuentry);
  string ssz=args[2]->soutput(menuentry);

  if ( !s.length() )
    return o << s;

  ps = atoi( sps.c_str() );
  if ( ps < 0 )
    ps = 0;
  if ( ps >= s.length() )
    ps = s.length() - 1;

  sz = atoi( ssz.c_str() );
  if ( sz < 0 )
    sz = 0;
  if ( ps + sz > s.length() )
    sz = s.length() - ps;

  return o << s.substr( ps, sz );
}

ostream &locate_icon::output(ostream &o, vector<cat_str *> &args,
    map<string, string> &menuentry)
{
  string path;
  string str_sz = args[0]->soutput(menuentry);
  string icon_name = args[1]->soutput(menuentry);
  std::list<string> paths;
  std::list<string> exts;
  int icon_name_sz = icon_name.size();

  if( str_sz.size() <= 0 )
  {
    str_sz = "16";
  }

  if( icon_name_sz <= 0 )
  {
    icon_name_sz = 15;
    icon_name = "default_section";
  }
  else if( icon_name.substr(0,1) == "/" )
  {
    return o << icon_name;
  }

  if( icon_name.substr(icon_name_sz-4,1) == "." )
  {
    icon_name_sz = icon_name_sz-4;
    icon_name = icon_name.substr(0, icon_name_sz);
  }

  exts.push_back(".png");
  exts.push_back(".xpm");

  paths.push_back(string("/usr/share/icons/Tango/") + str_sz +"x" + str_sz + "/apps/");
  paths.push_back(string("/usr/share/icons/TangoKDE/") + str_sz +"x" + str_sz + "/apps/");
  paths.push_back(string("/usr/share/icons/hicolor/") + str_sz +"x" + str_sz + "/apps/");
  paths.push_back(string("/usr/share/icons/default.kde/") + str_sz +"x" + str_sz + "/apps/");
  paths.push_back(string("/usr/share/icons/default.gnome/") + str_sz +"x" + str_sz + "/apps/");
  if( str_sz == "48" )
    paths.push_back("/usr/share/icons/large/");
  else if( str_sz == "32" )
    paths.push_back("/usr/share/icons/");
  else if( str_sz == "16" )
    paths.push_back("/usr/share/icons/mini/");
  paths.push_back("/usr/share/pixmaps/");

  std::list<string>::iterator it_ext;
  for(it_ext = exts.begin(); it_ext != exts.end(); it_ext++)
  {
    std::list<string>::iterator it_path;
    for(it_path = paths.begin(); it_path != paths.end(); it_path++)
    {
	string xpath(*it_path + icon_name + *it_ext);
	if( ::access(xpath.c_str(), F_OK) == 0 )
	{
	    path = xpath;
	    break;
	}
    }
  }

  if( path.size() <= 0 )
    path = string("/usr/share/icons/hicolor/") + str_sz +"x" + str_sz + "/apps/default_section.png";

  return o << path;
}

}
