#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

#include <cstring>
#include <cstdlib>
#include <iostream>
#include <fstream>

#include "common.h"
#include "xdg_categories_translator.h"

XdgCtgTranslator::XdgCtgTranslator()
{
    load(MENUMETHODS"/translate_xdg_categories");
    load(MENUMETHODS"/translate_menu_sections", true);
}

XdgCtgTranslator::~XdgCtgTranslator(){}

void XdgCtgTranslator::load(const std::string &translation_file, bool is_sections)
{
    FILE *transfile = fopen(translation_file.c_str(), "r");
    if(!transfile) return;
    while(!feof(transfile))
    {
	char tmp[MAX_LINE];
	std::string line;
	struct XdgCtgCondition condition;

	if( !fgets(tmp, MAX_LINE, transfile) ) continue;
	if( tmp[0] == '#' )
	    continue;
	int len = strlen(tmp);
	if( len < 5 )
	    continue;
	if( tmp[len-1] == '\n' )
	    line = std::string(tmp, 0, len-1);
	else
	    line = tmp;

	// take section name
	int eq_pos = line.find('=');
	if( eq_pos < 2 )
	    continue;
	condition.section = std::string(line, 0, eq_pos);
	line = std::string(line, eq_pos+1, line.size()-eq_pos-1);
	//std::cerr << "section<" << condition.section << "> line<" << line << ">" << std::endl;

	if(is_sections)
	{
	    section_to_category_map[condition.section] = line;
	}
	else
	{
	    // take conditions
	    std::list<std::string> condlist = split(line, ';');
	    for(std::list<std::string>::iterator it = condlist.begin(); it != condlist.end(); ++it)
	    {
		std::string subcondition = *it;
		if( subcondition[0] == '!' )
		{
		    condition.conflicts.push_back(subcondition.substr(1, subcondition.size()-1));
		    //std::cerr << "  conflict<" << subcondition.substr(1, subcondition.size()-1) << ">" << std::endl;
		}
		else
		{
		    condition.requires.push_back(subcondition);
		    //std::cerr << "  require<" << subcondition << ">" << std::endl;
		}
	    }
	    // add condition to list
	    if( condition.requires.size() > 0 || condition.conflicts.size() > 0 )
	    {
		conditions.push_back(condition);
	    }
	}
    }
}

std::string XdgCtgTranslator::translate(const std::string &categories)
{
    std::string section;
    std::list<std::string> ctglist = split(categories, ';');
    for(std::list<XdgCtgCondition>::iterator cond_it = conditions.begin(); cond_it != conditions.end(); ++cond_it)
    {
	XdgCtgCondition cond = *cond_it;
//	find conflicts
	unsigned int conflicts_found = 0;
	for(std::list<std::string>::iterator ctg_it = ctglist.begin(); ctg_it != ctglist.end(); ++ctg_it)
        {
	    std::string category = *ctg_it;
	    for(std::list<std::string>::iterator conflict_it = cond.conflicts.begin();
		conflict_it != cond.conflicts.end(); ++conflict_it)
	    {
		if( category == *conflict_it )
		{
		    conflicts_found++;
		    break;
		}
	    }
	}
//	find requires
	unsigned int requires_found = 0;
	for(std::list<std::string>::iterator ctg_it = ctglist.begin(); ctg_it != ctglist.end(); ++ctg_it)
        {
	    std::string category = *ctg_it;
	    for(std::list<std::string>::iterator require_it = cond.requires.begin();
		require_it != cond.requires.end(); ++require_it)
	    {
		if( category == *require_it )
		{
		    requires_found++;
		    break;
		}
	    }
	}
	if( cond.requires.size() == requires_found && cond.conflicts.size() == conflicts_found )
	{
	    section = cond.section;
	    break;
	}
    }
    return section;
}

std::string XdgCtgTranslator::translate_section(const std::string &section)
{
    std::string categories = section_to_category_map[section];
    if( categories.size() <= 0 )
    {
	categories = "UnknownCategory;";
    }
    return categories;
}

std::list<std::string> XdgCtgTranslator::split(std::string textline, const char separator)
{
    std::list<std::string> newlist;
    for(std::string::iterator it = textline.begin(); it != textline.end(); ++it)
    {
	std::string substring;
	while( it != textline.end() && *it != separator )
	{
	    substring += *it;
	    ++it;
	}
	if( it == textline.end() ) --it;
	if( substring.size() > 0 )
	    newlist.push_back(substring);
    }
    return newlist;
}

void XdgCtgTranslator::saveI18n()
{
    const char* home_dir;
    struct passwd *pwentry = getpwuid(getuid());
    if (pwentry != NULL)
	home_dir = pwentry->pw_dir;
    else
	home_dir = getenv("HOME");

    std::ofstream i18n_file((std::string(home_dir)+"/.menu-updates.i18n").c_str(), std::ios::trunc);

	for(std::map<std::string, std::map<std::string,std::string> >::iterator it = i18n.begin(); it != i18n.end(); ++it)
	{
	    std::string lng = (*it).first;
	    std::map<std::string, std::string> lng_strings = (*it).second;
	    for(std::map<std::string, std::string>::iterator it2 = lng_strings.begin(); it2 != lng_strings.end(); ++it2)
	    {
		i18n_file << lng << "|" << ((*it2).first).size() << "|" << (*it2).first << (*it2).second << std::endl;
	    }
	}
}

void XdgCtgTranslator::addI18n(const std::string &str)
{
}
